@@ -40,10 +40,6 @@
 //	internal routine prototypes
 
 static	int	RcdValErr( SGlobPtr GPtr, OSErr type, UInt32 correct, UInt32 incorrect, HFSCatalogNodeID parid );
-
-static	int	RcdFThdErr( SGlobPtr GPtr, UInt32 fid );
-
-static	int	RcdNoDirErr( SGlobPtr GPtr, UInt32 did );
 		
 static	int	RcdNameLockedErr( SGlobPtr GPtr, OSErr type, UInt32 incorrect );
 	
@@ -65,6 +61,77 @@ static OSErr  GetVolumeHeaderBlock(SVCB *vcb, HFSMasterDirectoryBlock *mdb, Bloc
 OSErr	ScavengeVolumeType( SGlobPtr GPtr, HFSMasterDirectoryBlock *mdb, UInt16 *volumeType );
 OSErr	SeekVolumeHeader( SGlobPtr GPtr, UInt64 startSector, UInt32 numSectors, UInt64 *vHSector );
 
+/*
+ * Check if a volume is journaled.  
+ *
+ * returns:    	0 not journaled
+ *		1 journaled
+ *
+ */
+int
+CheckIfJournaled(SGlobPtr GPtr)
+{
+#define kIDSector 2
+
+	OSErr err;
+	int result;
+	HFSMasterDirectoryBlock	*mdbp;
+	HFSPlusVolumeHeader *vhp;
+	SVCB *vcb = GPtr->calculatedVCB;
+	ReleaseBlockOptions rbOptions;
+	BlockDescriptor block;
+
+	vhp = (HFSPlusVolumeHeader *) NULL;
+	rbOptions = kReleaseBlock;
+
+	err = GetVolumeBlock(vcb, kIDSector, kGetBlock, &block);
+	if (err) return (0);
+
+	mdbp = (HFSMasterDirectoryBlock	*) block.buffer;
+
+	if (mdbp->drSigWord == kHFSPlusSigWord) {
+		vhp = (HFSPlusVolumeHeader *) block.buffer;
+
+	} else if (mdbp->drSigWord == kHFSSigWord) {
+
+		if (mdbp->drEmbedSigWord == kHFSPlusSigWord) {
+			UInt32 vhSector;
+			UInt32 blkSectors;
+			
+			blkSectors = mdbp->drAlBlkSiz / 512;
+			vhSector  = mdbp->drAlBlSt;
+			vhSector += blkSectors * mdbp->drEmbedExtent.startBlock;
+			vhSector += kIDSector;
+	
+			(void) ReleaseVolumeBlock(vcb, &block, kReleaseBlock);
+			err = GetVolumeBlock(vcb, vhSector, kGetBlock, &block);
+			if (err) return (0);
+
+			vhp = (HFSPlusVolumeHeader *) block.buffer;
+			mdbp = (HFSMasterDirectoryBlock	*) NULL;
+
+		}
+	}
+
+	if ((vhp != NULL) && (ValidVolumeHeader(vhp) == noErr)) {
+		result = ((vhp->attributes & kHFSVolumeJournaledMask) != 0);
+
+		// even if journaling is enabled for this volume, we'll return
+		// false if it wasn't unmounted cleanly and it was previously
+		// mounted by someone that doesn't know about journaling.
+		if (   vhp->lastMountedVersion != kHFSJMountVersion
+			&& (vhp->attributes & kHFSVolumeUnmountedMask) == 0) {
+			result = 0;
+		}
+	} else {
+		result = 0;
+	}
+
+	(void) ReleaseVolumeBlock(vcb, &block, rbOptions);
+
+	return (result);
+}
+
 /*
  * Check if a volume is clean (unmounted safely)
  *
@@ -139,7 +206,6 @@ CheckForClean(SGlobPtr GPtr, Boolean markClean)
 	return (result);
 }
 
-
 /*------------------------------------------------------------------------------
 
 Function:	IVChk - (Initial Volume Check)
@@ -174,7 +240,7 @@ OSErr IVChk( SGlobPtr GPtr )
 	UInt32					i;
 	UInt32					hfsPlusIOPosOffset;
 	BTreeControlBlock		*btcb;
-	SVCB				*vcb	= GPtr->calculatedVCB;
+	SVCB					*vcb	= GPtr->calculatedVCB;
 	
 	//  Set up
 
@@ -349,7 +415,7 @@ again:
 	
 	vcb->vcbTotalBlocks	= realTotalBlocks;
 	vcb->vcbFreeBlocks	= 0;
-	//	Only do these tests on HFS volumes, since they are either irrellivent
+	//	Only do these tests on HFS volumes, since they are either 
 	//	or, getting the VolumeHeader would have already failed.
 
 	if ( GPtr->isHFSPlus == false )
@@ -643,8 +709,9 @@ OSErr	CreateExtentsBTreeControlBlock( SGlobPtr GPtr )
 		}
 		else
 		{
-			GPtr->calculatedExtentsFCB->fcbLogicalSize  = (UInt32) volumeHeader->extentsFile.logicalSize;					//	Set Extents tree's LEOF
-			GPtr->calculatedExtentsFCB->fcbPhysicalSize = volumeHeader->extentsFile.totalBlocks * volumeHeader->blockSize;	//	Set Extents tree's PEOF
+			GPtr->calculatedExtentsFCB->fcbLogicalSize  = volumeHeader->extentsFile.logicalSize;					//	Set Extents tree's LEOF
+			GPtr->calculatedExtentsFCB->fcbPhysicalSize = (UInt64)volumeHeader->extentsFile.totalBlocks * 
+														  (UInt64)volumeHeader->blockSize;	//	Set Extents tree's PEOF
 		}
 
 		//
@@ -695,7 +762,7 @@ OSErr	CreateExtentsBTreeControlBlock( SGlobPtr GPtr )
 		err = CheckFileExtents( GPtr, kHFSExtentsFileID, 0, (void *)GPtr->calculatedExtentsFCB->fcbExtents16, &numABlks );	/* check out extent info */	
 		if (err) goto exit;
 	
-		if (alternateMDB->drXTFlSize != (numABlks * GPtr->calculatedVCB->vcbBlockSize))//	check out the PEOF
+		if (alternateMDB->drXTFlSize != ((UInt64)numABlks * (UInt64)GPtr->calculatedVCB->vcbBlockSize))//	check out the PEOF
 		{
 			RcdError(GPtr,E_ExtPEOF);
 			err = E_ExtPEOF;
@@ -956,8 +1023,9 @@ OSErr	CreateCatalogBTreeControlBlock( SGlobPtr GPtr )
 		}
 		else
 		{
-			GPtr->calculatedCatalogFCB->fcbLogicalSize  = (UInt32) volumeHeader->catalogFile.logicalSize;
-			GPtr->calculatedCatalogFCB->fcbPhysicalSize = volumeHeader->catalogFile.totalBlocks * volumeHeader->blockSize;
+			GPtr->calculatedCatalogFCB->fcbLogicalSize  = volumeHeader->catalogFile.logicalSize;
+			GPtr->calculatedCatalogFCB->fcbPhysicalSize = (UInt64)volumeHeader->catalogFile.totalBlocks * 
+														  (UInt64)volumeHeader->blockSize;  
 		}
 
 		//
@@ -1008,7 +1076,7 @@ OSErr	CreateCatalogBTreeControlBlock( SGlobPtr GPtr )
 		err = CheckFileExtents( GPtr, kHFSCatalogFileID, 0, (void *)GPtr->calculatedCatalogFCB->fcbExtents16, &numABlks );	/* check out extent info */	
 		if (err) goto exit;
 
-		if (alternateMDB->drCTFlSize != (numABlks * vcb->vcbBlockSize))	//	check out the PEOF
+		if (alternateMDB->drCTFlSize != ((UInt64)numABlks * (UInt64)vcb->vcbBlockSize))	//	check out the PEOF
 		{
 			RcdError( GPtr, E_CatPEOF );
 			err = E_CatPEOF;
@@ -1133,8 +1201,9 @@ OSErr	CreateExtendedAllocationsFCB( SGlobPtr GPtr )
 		}
 		else
 		{
-			fcb->fcbLogicalSize  = (UInt32) volumeHeader->allocationFile.logicalSize;
-			fcb->fcbPhysicalSize = volumeHeader->allocationFile.totalBlocks * volumeHeader->blockSize;
+			fcb->fcbLogicalSize  = volumeHeader->allocationFile.logicalSize;
+			fcb->fcbPhysicalSize = (UInt64) volumeHeader->allocationFile.totalBlocks * 
+								   (UInt64) volumeHeader->blockSize; 
 		}
 
 		/* while we're here, also get startup file extents... */
@@ -1144,8 +1213,9 @@ OSErr	CreateExtendedAllocationsFCB( SGlobPtr GPtr )
 		err = CheckFileExtents( GPtr, kHFSStartupFileID, 0, (void *)fcb->fcbExtents32, &numABlks );
 		if (err) goto exit;
 
-		fcb->fcbLogicalSize  = (UInt32) volumeHeader->startupFile.logicalSize;
-		fcb->fcbPhysicalSize = volumeHeader->startupFile.totalBlocks * volumeHeader->blockSize;
+		fcb->fcbLogicalSize  = volumeHeader->startupFile.logicalSize;
+		fcb->fcbPhysicalSize = (UInt64) volumeHeader->startupFile.totalBlocks * 
+								(UInt64) volumeHeader->blockSize; 
 	}
 
 exit:
@@ -1574,7 +1644,8 @@ OSErr	CreateAttributesBTreeControlBlock( SGlobPtr GPtr )
 		else
 		{
 			GPtr->calculatedAttributesFCB->fcbLogicalSize  = (UInt64) volumeHeader->attributesFile.logicalSize;						//	Set Attributes tree's LEOF
-			GPtr->calculatedAttributesFCB->fcbPhysicalSize = volumeHeader->attributesFile.totalBlocks * volumeHeader->blockSize;	//	Set Attributes tree's PEOF
+			GPtr->calculatedAttributesFCB->fcbPhysicalSize = (UInt64) volumeHeader->attributesFile.totalBlocks * 
+											(UInt64) volumeHeader->blockSize;	//	Set Attributes tree's PEOF
 		}
 
 		//
@@ -1747,6 +1818,7 @@ OSErr AttrBTChk( SGlobPtr GPtr )
 
 
 
+#if 0
 /*------------------------------------------------------------------------------
 
 Name:		RcdFThdErr - (record file thread error)
@@ -1814,6 +1886,7 @@ static int RcdNoDirErr( SGlobPtr GPtr, UInt32 did )			//	the directory ID in the
 	
 	return( noErr );										//	successful return
 }
+#endif
 
 
 /*------------------------------------------------------------------------------
@@ -2389,7 +2462,8 @@ Output:
 			blocksUsed	-	number of allocation blocks allocated to the file
 ------------------------------------------------------------------------------*/
 
-OSErr	CheckFileExtents( SGlobPtr GPtr, UInt32 fileNumber, UInt8 forkType, const void *extents, UInt32 *blocksUsed )
+OSErr	CheckFileExtents( SGlobPtr GPtr, UInt32 fileNumber, UInt8 forkType, 
+						  const void *extents, UInt32 *blocksUsed )
 {
 	UInt32				blockCount;
 	UInt32				extentBlockCount;
